# coding: utf-8
#
# This code is part of pyrekordbox.
#
# Copyright (c) 2022, Dylan Jones

"""Rekordbox XML database file interface."""

import logging
import os.path
from xml.dom import minidom
import xml.etree.cElementTree as xml

logger = logging.getLogger(__name__)


def pretty_xml(element, indent=None, encoding="utf-8"):
    """Generates a formatted string of an XML element.

    Parameters
    ----------
    element : xml.etree.cElementTree.Element
        The input XML element.
    indent : str, optional
        The indentation used for formatting the XML element. The default is 3 spaces.
    encoding : str, optional
        The encoding used for formatting the XML element. The default is `utf-8`.

    Returns
    -------
    xml_string : str
        The formatted string of the XML element.
    """
    # Build pretty xml-string
    if indent is None:
        indent = "    "
    if encoding is None:
        encoding = "utf-8"
    rough_string = xml.tostring(element, encoding)
    reparsed = minidom.parseString(rough_string)
    string = reparsed.toprettyxml(indent=indent, encoding=encoding).decode()
    # Remove annoying empty lines
    string = "\n".join([line for line in string.splitlines() if line.strip()])
    return string


# ======================================================================================
# XML elements
# ======================================================================================


class AbstractElement:
    """Abstract base class for Rekordbox XML elements."""

    TAG: str

    def __init__(self, element):
        self.element = element

    @property
    def attrib(self):
        return self.element.attrib

    def __repr__(self):
        return f"{self.__class__.__name__}({self.attrib})"


# -- Collection elements ---------------------------------------------------------------


class Tempo(AbstractElement):
    """Rekordbox XML Tempo element for storing tempo and beat grid data of a track."""

    TAG = "TEMPO"

    def __init__(self, element):
        super().__init__(element)

    @classmethod
    def new(cls, track_element, inizio, bpm, metro, battito):
        """Create a new Tempo element

        Parameters
        ----------
        track_element: xml.etree.ElementTree.Element
            Parent track xml-element
        inizio: float
            Start position of BeatGrid
        bpm: float
            Value of BPM
        metro: str
            Kind of musical meter
        battito: int
            Beat number in the bar

        Returns
        -------
        temp: Tempo
        """
        attrib = {
            "Inizio": str(inizio),
            "Bpm": str(bpm),
            "Metro": metro,
            "Battito": str(battito),
        }
        return cls(xml.SubElement(track_element, cls.TAG, attrib=attrib))

    @property
    def inizio(self):
        """float: Start position of BeatGrid."""
        return float(self.element.get("Inizio"))

    @property
    def bpm(self):
        """float: Value of BPM."""
        return float(self.element.get("Bpm"))

    @property
    def metro(self):
        """str: Kind of musical meter."""
        return self.element.attrib.get("Metro")

    @property
    def battito(self):
        """int: Beat number in the bar."""
        return int(self.element.attrib.get("Battito"))


class PositionMark(AbstractElement):
    """Rekordbox XML Position element for storing position markers like cue points."""

    TAG = "POSITION_MARK"

    CUE = 0
    FADE_IN = 1
    FADE_OUT = 2
    LOAD = 3
    LOOP = 4

    def __init__(self, element):
        super().__init__(element)

    @classmethod
    def new(cls, track_element, name, marker_type, start, end, num):
        """Create a new PositionMark element

        Parameters
        ----------
        track_element : xml.etree.ElementTree.Element
            Parent track xml-element.
        name : str
            Name of position mark.
        marker_type : int
            Type of position mark.
        start : float
            Start position of position mark.
        end : float
            End position of position mark.
        num : int
            Number for identification of the position mark.

        Returns
        -------
        mark : PositionMark
            The new ``PositionMark`` instance.
        """
        attrib = {
            "Name": name,
            "Type": str(marker_type),
            "Start": str(start),
            "End": str(end),
            "Num": str(num),
        }
        return cls(xml.SubElement(track_element, cls.TAG, attrib=attrib))

    @property
    def name(self):
        """str: Name of position mark."""
        return self.element.attrib.get("Name")

    @property
    def type(self):
        """int: Type of position mark (Cue=0, Fade-In=1, Fade-Out=2, Load=3, Loop=4)."""
        return int(self.element.attrib.get("Type"))

    @property
    def start(self):
        """float: Start position of position mark."""
        return float(self.element.attrib.get("Start"))

    @property
    def end(self):
        """float: End position of position mark."""
        val = self.element.attrib.get("End")
        return float(val) if val is not None else None

    @property
    def num(self):
        """int: Number for identification of the position mark.

        Hot-Cue: 0, 1, 2, ...; Memory: -1.
        """
        return int(self.element.attrib.get("Num"))


class Track(AbstractElement):
    """Rekordbox XML Track element for storing the metadata of a track."""

    TAG = "TRACK"

    ATTRIBS = [
        "TrackID",
        "Name",
        "Artist",
        "Composer",
        "Album",
        "Grouping",
        "Genre",
        "Kind",
        "Size",
        "TotalTime",
        "DiscNumber",
        "TrackNumber",
        "Year",
        "AverageBpm",
        "DateModified",
        "DateAdded",
        "BitRate",
        "SampleRate",
        "Comments",
        "PlayCount",
        "LastPlayed",
        "Rating",
        "Location",
        "Remixer",
        "Tonality",
        "Label",
        "Mix",
        "Colour",
    ]

    def __init__(self, element):
        super().__init__(element)
        self.tempos = list()
        self.marks = list()
        self.load_subelements()

    def load_subelements(self):
        tempo_elements = self.element.findall(f"{Tempo.TAG}")
        if tempo_elements is not None:
            self.tempos = [Tempo(el) for el in tempo_elements]
        mark_elements = self.element.findall(f".//{PositionMark.TAG}")
        if mark_elements is not None:
            self.marks = [PositionMark(el) for el in mark_elements]

    @property
    def attrib(self):
        return self.element.attrib

    @property
    def id(self):
        """int: Identification of track."""
        return int(self.element.attrib.get("TrackID"))

    @id.setter
    def id(self, value):
        """int: Identification of track."""
        self.element.attrib.update({"TrackID": str(value)})

    @property
    def name(self):
        """str: Name of track."""
        return self.element.attrib.get("Name")

    @name.setter
    def name(self, value):
        """str: Name of track."""
        self.element.attrib.update({"Name": value})

    @property
    def artist(self):
        """str: Name of artist."""
        return self.element.attrib.get("Artist")

    @artist.setter
    def artist(self, value):
        """str: Name of artist."""
        self.element.attrib.update({"Artist": value})

    @property
    def composer(self):
        """str: Name of composer (or producer)."""
        return self.element.attrib.get("Composer")

    @composer.setter
    def composer(self, value):
        """str: Name of composer (or producer)"""
        self.element.attrib.update({"Composer": value})

    @property
    def album(self):
        """str: Name of album."""
        return self.element.attrib.get("Album")

    @album.setter
    def album(self, value):
        """str: Name of album."""
        self.element.attrib.update({"Album": value})

    @property
    def grouping(self):
        """str: Name of group."""
        return self.element.attrib.get("Grouping")

    @grouping.setter
    def grouping(self, value):
        """str: Name of group."""
        self.element.attrib.update({"Grouping": value})

    @property
    def genre(self):
        """str: Name of genre."""
        return self.element.attrib.get("Genre")

    @genre.setter
    def genre(self, value):
        """str: Name of genre."""
        self.element.attrib.update({"Genre": value})

    @property
    def kind(self):
        """str: Type of audio file."""
        return self.element.attrib.get("Kind")

    @kind.setter
    def kind(self, value):
        """str: Type of audio file."""
        self.element.attrib.update({"Kind": value})

    @property
    def size(self):
        """int: Size of audio file."""
        return int(self.element.attrib.get("Size"))

    @size.setter
    def size(self, value):
        """int: Size of audio file."""
        self.element.attrib.update({"Size": str(value)})

    @property
    def total_time(self):
        """float: Duration of track (in seconds)."""
        return float(self.element.attrib.get("TotalTime"))

    @total_time.setter
    def total_time(self, value):
        """float: Duration of track (in seconds)."""
        self.element.attrib.update({"TotalTime": str(value)})

    @property
    def disc_num(self):
        """int: Order number of the disc of the album."""
        return int(self.element.attrib.get("DiscNumber"))

    @disc_num.setter
    def disc_num(self, value):
        """int: Order number of the disc of the album."""
        self.element.attrib.update({"DiscNumber": value})

    @property
    def track_num(self):
        """int: Order number of the track on the album."""
        return int(self.element.attrib.get("TrackNumber"))

    @track_num.setter
    def track_num(self, value):
        """int: Order number of the track on the album."""
        self.element.attrib.update({"TrackNumber": value})

    @property
    def year(self):
        """str: Year of release."""
        return self.element.attrib.get("Year")

    @year.setter
    def year(self, value):
        """str: Year of release."""
        self.element.attrib.update({"Year": value})

    @property
    def bpm(self):
        """float Value of average BPM."""
        return float(self.element.attrib.get("AverageBPM"))

    @bpm.setter
    def bpm(self, value):
        """float Value of average BPM."""
        self.element.attrib.update({"AverageBPM": str(value)})

    @property
    def date_modified(self):
        """str: Date of last modification (Format: yyyy-mm-dd)."""
        return self.element.attrib.get("DateModified")

    @date_modified.setter
    def date_modified(self, value):
        """str: Date of last modification (Format: yyyy-mm-dd)."""
        self.element.attrib.update({"DateModified": value})

    @property
    def date_added(self):
        """str: Date of addition (Format: yyyy-mm-dd)."""
        return self.element.attrib.get("DateAdded")

    @date_added.setter
    def date_added(self, value):
        """str: Date of addition (Format: yyyy-mm-dd)."""
        self.element.attrib.update({"DateAdded": value})

    @property
    def bit_rate(self):
        """int: Encoding bit rate."""
        return int(self.element.attrib.get("BitRate"))

    @bit_rate.setter
    def bit_rate(self, value):
        """int: Encoding bit rate."""
        self.element.attrib.update({"BitRate": str(value)})

    @property
    def sample_rate(self):
        """int: Frequency of sampling."""
        return int(self.element.attrib.get("SampleRate"))

    @sample_rate.setter
    def sample_rate(self, value):
        """int: Frequency of sampling."""
        self.element.attrib.update({"SampleRate": str(value)})

    @property
    def comments(self):
        """str: Comments of track."""
        return self.element.attrib.get("Comments")

    @comments.setter
    def comments(self, value):
        """str: Comments of track."""
        self.element.attrib.update({"Comments": value})

    @property
    def play_count(self):
        """int: Play count of the track."""
        return int(self.element.attrib.get("PlayCount"))

    @play_count.setter
    def play_count(self, value):
        """int: Play count of the track."""
        self.element.attrib.update({"PlayCount": str(value)})

    @property
    def last_played(self):
        """str: Date of last playing (Format: yyyy-mm-dd)."""
        return self.element.attrib.get("LastPlayed")

    @last_played.setter
    def last_played(self, value):
        """str: Date of last playing (Format: yyyy-mm-dd)."""
        self.element.attrib.update({"LastPlayed": value})

    @property
    def rating(self):
        """int: Rating of the track.

        0 star=0, 1 star=51, 2 stars=102, 3 stars=153, 4 stars=204, 5 stars=255.
        """
        return int(self.element.attrib.get("Rating"))

    @rating.setter
    def rating(self, value):
        """int: Rating of the track.

        0 star=0, 1 star=51, 2 stars=102, 3 stars=153, 4 stars=204, 5 stars=255.
        """
        self.element.attrib.update({"Rating": str(value)})

    @property
    def location(self):
        """str: Location of the file."""
        return self.element.attrib.get("Location")

    @location.setter
    def location(self, value):
        """str: Location of the file."""
        self.element.attrib.update({"Location": value})

    @property
    def remixer(self):
        """str: Name of remixer."""
        return self.element.attrib.get("Remixer")

    @remixer.setter
    def remixer(self, value):
        """str: Name of remixer."""
        self.element.attrib.update({"Remixer": value})

    @property
    def key(self):
        """str: Tonality (Kind of musical key)."""
        return self.element.attrib.get("Tonality")

    @key.setter
    def key(self, value):
        """str: Tonality (Kind of musical key)."""
        self.element.attrib.update({"Tonality": value})

    @property
    def label(self):
        """str: Name of the record label."""
        return self.element.attrib.get("Label")

    @label.setter
    def label(self, value):
        """str: Name of the record label."""
        self.element.attrib.update({"Label": value})

    @property
    def mix(self):
        """str: Name of mix."""
        return self.element.attrib.get("Mix")

    @mix.setter
    def mix(self, value):
        """str: Name of mix."""
        self.element.attrib.update({"Mix": value})

    @property
    def color(self):
        """str: Colour for track grouping."""
        return self.element.attrib.get("Colour")

    @color.setter
    def color(self, value):
        """str: Colour for track grouping."""
        self.element.attrib.update({"Colour": value})

    def sort_markers(self):
        self.marks.sort(key=lambda x: x.start)

    def cue_points(self):
        return [m for m in self.marks if m.type == PositionMark.CUE]

    def to_string(self):
        string = "TRACK:\n"
        for key, value in self.attrib.items():
            string += f"{key:<13}{value}\n"
        return string

    def __repr__(self):
        track_id = self.id
        name = self.name
        return f"{self.__class__.__name__}(ID={track_id}, Name={name})"


# -- Playlist elements -----------------------------------------------------------------


class Node(AbstractElement):
    """Rekordbox XML Node element used for storing playlist folders and playlists."""

    TAG = "NODE"

    FOLDER = 0
    PLAYLIST = 1

    TRACKID = 0
    LOCATION = 1

    def __init__(self, element):
        super().__init__(element)

    @classmethod
    def folder(cls, parent_element, name):
        attrib = {"Name": name, "Type": str(cls.FOLDER), "Count": "0"}
        return cls(xml.SubElement(parent_element, cls.TAG, attrib=attrib))

    @classmethod
    def playlist(cls, parent_element, name, keytype=TRACKID):
        attrib = {
            "Name": name,
            "Type": str(cls.PLAYLIST),
            "KeyType": str(keytype),
            "Entries": "0",
        }
        return cls(xml.SubElement(parent_element, cls.TAG, attrib=attrib))

    @property
    def parent(self):
        return self.element.parent

    @property
    def name(self):
        """str: Name of Node."""
        return self.element.attrib.get("Name")

    @property
    def type(self):
        """int: Type of Node (0=folder or 1=playlist)."""
        return int(self.element.attrib.get("Type"))

    @property
    def count(self):
        """int: Number of Node's in the Node."""
        return int(self.element.attrib.get("Count", 0))

    @property
    def entries(self):
        """int: Number of Track's in Playlist."""
        return int(self.element.attrib.get("Entries", 0))

    @property
    def keytype(self):
        """int: Kind of identification (0=TrackID or 1=Location)."""
        return int(self.element.attrib.get("KeyType"))

    @property
    def keys(self):
        """list of Track: TrackID's of the playlist-content (if playlist-Node)."""
        keys = list()
        for el in self.element.findall(f".//{Track.TAG}"):
            val = el.attrib["Key"]
            keys.append(int(val) if self.keytype == self.TRACKID else val)
        return keys

    @property
    def nodes(self):
        """list of Node: Sub-Nodes (in case of a folder-Node)."""
        return [Node(el) for el in self.element.findall(f"{Node.TAG}")]

    def get_node(self, i):
        """Returns the i-th sub-Node of the current node.

        Parameters
        ----------
        i : int
            Index of sub-Node

        Returns
        -------
        subnode : Node
        """
        return Node(self.element.findall(f"{Node.TAG}[{i}]"))

    def get_node_by_name(self, name):
        """Returns the sub-Node with the given name.

        Parameters
        ----------
        name : str
            Name of the sub-Node

        Returns
        -------
        subnode : Node
        """
        return Node(self.element.find(f'.//{Node.TAG}[@Name="{name}"]'))

    def add_folder_node(self, name):
        node = Node.folder(self.element, name)
        self.element.attrib["Count"] = str(self.count + 1)
        return node

    def add_playlist_node(self, name, keytype=TRACKID):
        node = Node.playlist(self.element, name, keytype)
        self.element.attrib["Count"] = str(self.count + 1)
        return node

    def get_track(self, idx):
        if self.type == self.FOLDER:
            return None
        el = self.element.findall(f".//{Track.TAG}[{idx}]")
        return int(el.attrib["Key"])

    def add_track(self, key):
        """Add a track to the current node.

        Parameters
        ----------
        key : int or str
            Track identification.
        """
        el = xml.SubElement(self.element, Track.TAG, attrib={"Key": str(key)})
        self.element.attrib["Entries"] = str(self.entries + 1)
        return el

    def print_tree(self, lvl=0, indent=4):
        space = indent * lvl * " "
        if self.type == self.PLAYLIST:
            print(space + f"Playlist: {self.name} ({self.entries} Tracks)")
        elif self.type == self.FOLDER:
            print(space + f"Folder: {self.name}")
            for node in self.nodes:
                node.print_tree(lvl + 1, indent)

    def __repr__(self):
        string = f"Node(Name={self.name}, type={self.type}, "
        if self.type == self.FOLDER:
            string += f"Count={self.count}"
        else:
            string += f"Entries={self.entries}, KeyType={self.keytype}"
        return string + ")"


# ======================================================================================
# XML File
# ======================================================================================


class RekordboxXml:
    """Rekordbox XML database file object."""

    ROOT = "DJ_PLAYLISTS"
    PRDT_TAG = "PRODUCT"
    PLST_TAG = "PLAYLISTS"
    COLL_TAG = "COLLECTION"

    def __init__(self, path="", name=None, version=None, company=None):
        self.path = path
        self._root = None
        self._product = None
        self._collection = None
        self._playlists = None
        self._root_node = None

        self.collection = None

        if os.path.exists(path):
            self.parse(self.path)
        else:
            self.init(name, version, company)

    @property
    def frmt_version(self):
        return self._root.attrib["Version"]

    @property
    def product_name(self):
        return self._product.attrib.get("Name")

    @property
    def product_version(self):
        return self._product.attrib.get("Version")

    @property
    def product_company(self):
        return self._product.attrib.get("Company")

    @property
    def num_tracks(self):
        return int(self._collection.attrib.get("Entries"))

    def init(self, name=None, version=None, company=None, frmt_version=None):
        frmt_version = frmt_version or "1.0.0"
        name = name or "pyRekordbox"
        version = version or "0.0.1"
        company = company or "None"

        # Initialize root element
        self._root = xml.Element(self.ROOT, attrib={"Version": frmt_version})
        # Initialize product element
        attrib = {"Name": name, "Version": version, "Company": company}
        self._product = xml.SubElement(self._root, self.PRDT_TAG, attrib=attrib)
        # Initialize collection element
        attrib = {"Entries": "0"}
        self._collection = xml.SubElement(self._root, self.COLL_TAG, attrib=attrib)
        # Initialize playlist element
        self._playlists = xml.SubElement(self._root, self.PLST_TAG)
        self._root_node = Node.folder(self._playlists, "ROOT")

    def parse(self, path):
        self.path = path

        tree = xml.parse(path)
        self._root = tree.getroot()
        self._product = self._root.find(self.PRDT_TAG)
        self._collection = self._root.find(self.COLL_TAG)
        self._playlists = self._root.find(self.PLST_TAG)
        self._root_node = Node(self._playlists.find(Node.TAG))

    def get_tracks(self):
        return [Track(el) for el in self._collection.findall(f".//{Track.TAG}")]

    def get_track_ids(self):
        return [Track(el).id for el in self._collection.findall(f".//{Track.TAG}")]

    def get_track(self, track_id):
        element = self._collection.find(f'.//{Track.TAG}[@TrackID="{track_id}"]')
        if element is None:
            return None
        return Track(element)

    def update_track_count(self):
        num_tracks = len(self._collection.findall(f".//{Track.TAG}"))
        self._collection.attrib["Entries"] = num_tracks

    def get_playlist(self, *names):
        node = self._root_node
        for name in names:
            node = node.get_node_by_name(name)
        return node

    def print_playlist_tree(self):
        self._root_node.print_tree()

    def tostring(self, indent=None):
        return pretty_xml(self._root, indent, encoding="utf-8")

    def save(self, path="", indent=None):
        if not path:
            path = self.path
        string = self.tostring(indent)
        with open(path, "w") as fh:
            fh.write(string)

    def __repr__(self):
        path = self.path
        name = self.product_name
        v = self.product_version
        company = self.product_company
        tracks = self.num_tracks

        cls = self.__class__.__name__
        s = f"{cls}('{path}', tracks={tracks}, info={name} {company} v{v})"
        return s



# coding: utf-8
#
# This code is part of pyrekordbox.
#
# Copyright (c) 2022, Dylan Jones

import logging
from abc import ABC
import numpy as np
from . import structs

logger = logging.getLogger(__name__)


class BuildTagLengthError(Exception):
    def __init__(self, struct, len_data):
        super().__init__(
            f"`len_tag` ({struct.len_tag}) of '{struct.type}' does not "
            f"match the data-length ({len_data})!"
        )


class AbstractAnlzTag(ABC):
    """Abstract base class for struct handlers of Rekordbox analysis files."""

    type: str
    name: str
    LEN_HEADER: int = 0  # Expected value of `len_header`
    LEN_TAG: int = 0  # Expected value of `len_tag`

    def __init__(self, tag_data):
        self.struct = None
        if tag_data is not None:
            self.parse(tag_data)

    @property
    def content(self):
        return self.struct.content

    def _check_len_header(self):
        if self.LEN_HEADER != self.struct.len_header:
            logger.warning(
                "`len_header` (%s) of `%s` doesn't match the expected value %s",
                self.struct.len_header,
                self.struct.type,
                self.LEN_HEADER,
            )

    def _check_len_tag(self):
        if self.LEN_TAG != self.struct.len_tag:
            logger.warning(
                "`len_tag` (%s)  of `%s` doesn't match the expected value %s",
                self.struct.len_tag,
                self.struct.type,
                self.LEN_TAG,
            )

    def check_parse(self):
        pass

    def parse(self, tag_data):
        self.struct = structs.AnlzTag.parse(tag_data)
        if self.LEN_HEADER:
            self._check_len_header()
        if self.LEN_TAG:
            self._check_len_tag()
        self.check_parse()

    def build(self):
        data = structs.AnlzTag.build(self.struct)
        len_data = len(data)
        if len_data != self.struct.len_tag:
            raise BuildTagLengthError(self.struct, len_data)
        return data

    def get(self):
        return self.struct.content

    def set(self, *args, **kwargs) -> bytes:
        pass

    def update_len(self):
        pass

    def __repr__(self):
        len_header = self.struct.len_header
        len_tag = self.struct.len_tag
        return f"{self.__class__.__name__}(len_header={len_header}, len_tag={len_tag})"

    def __str__(self):
        return str(self.struct)


def _parse_wf_preview(tag):
    n = len(tag.entries)
    wf = np.zeros(n)
    col = np.zeros(n)
    for i in range(n):
        data = tag.entries[i]
        wf[i] = data & 0x1F
        col[i] = data >> 5
    return wf, col


class PQTZAnlzTag(AbstractAnlzTag):
    """Beat grid struct handler."""

    type = "PQTZ"
    name = "beat_grid"
    LEN_HEADER = 24

    @property
    def beats(self):
        return self.get_beats()

    @property
    def bpms(self):
        return self.get_bpms()

    @property
    def bpms_average(self):
        if len(self.content.entries):
            return np.mean(self.get_bpms())
        return 0.0

    @property
    def bpms_unique(self):
        return np.unique(self.get_bpms())

    @property
    def times(self):
        return self.get_times()

    def check_parse(self):
        assert self.struct.content.entry_count == len(self.struct.content.entries)

    def get(self):
        n = len(self.content.entries)
        beats = np.zeros(n, dtype=np.int8)
        bpms = np.zeros(n, dtype=np.float64)
        times = np.zeros(n, dtype=np.float64)
        for i, entry in enumerate(self.content.entries):
            _, beat, bpm, time = entry.values()
            beats[i] = beat
            bpms[i] = bpm / 100  # BPM is saved as 100 * BPM
            times[i] = time / 1_000  # Convert milliseconds to seconds
        return beats, bpms, times

    def get_beats(self):
        return np.array([entry.beat for entry in self.content.entries], dtype=np.int8)

    def get_bpms(self):
        return np.array([entry.tempo / 100 for entry in self.content.entries])

    def get_times(self):
        return np.array([entry.time / 1000 for entry in self.content.entries])

    def set(self, beats, bpms, times):
        for i, (beat, bpm, t) in enumerate(zip(beats, bpms, times)):
            data = {"beat": int(beat), "tempo": int(100 * bpm), "time": int(1000 * t)}
            self.content.entries[i].update(data)

    def set_beats(self, beats):
        for i, beat in enumerate(beats):
            self.content.entries[i].beat = beat

    def set_bpms(self, bpms):
        for i, bpm in enumerate(bpms):
            self.content.entries[i].bpm = int(bpm * 100)

    def set_times(self, times):
        for i, t in enumerate(times):
            self.content.entries[i].time = int(1000 * t)

    def update_len(self):
        self.struct.len_tag = self.struct.len_header + 8 * len(self.content.entries)


class PQT2AnlzTag(AbstractAnlzTag):
    """Extended (nxs2) beat grid struct handler."""

    type = "PQT2"
    name = "beat_grid2"
    LEN_HEADER = 56

    @property
    def beats(self):
        return self.get_beats()

    @property
    def bpms(self):
        return self.get_bpms()

    @property
    def times(self):
        return self.get_times()

    @property
    def beat_grid_count(self):
        return self.content.entry_count

    @property
    def bpms_unique(self):
        return np.unique(self.get_bpms())

    def check_parse(self):
        len_beats = self.struct.content.entry_count
        if len_beats:
            expected = self.struct.len_tag - self.struct.len_header
            actual = 2 * len(self.content.entries)  # each entry consist of 2 bytes
            assert actual == expected, f"{actual} != {expected}"

    def get(self):
        n = len(self.content.entries)
        beats = np.zeros(n, dtype=np.int8)
        bpms = np.zeros(n, dtype=np.float64)
        times = np.zeros(n, dtype=np.float64)
        for i, entry in enumerate(self.content.entries):
            _, beat, bpm, time = entry.values()
            beats[i] = beat
            bpms[i] = bpm / 100  # BPM is saved as 100 * BPM
            times[i] = time / 1_000  # Convert milliseconds to seconds
        return beats, bpms, times

    def get_beats(self):
        return np.array([entry.beat for entry in self.content.entries], dtype=np.int8)

    def get_bpms(self):
        return np.array([entry.tempo / 100 for entry in self.content.entries])

    def get_times(self):
        return np.array([entry.time / 1000 for entry in self.content.entries])

    def set_beats(self, beats):
        for i, beat in enumerate(beats):
            self.content.entries[i].beat = beat

    def set_bpms(self, bpms):
        for i, bpm in enumerate(bpms):
            self.content.entries[i].bpm = int(bpm * 100)

    def set_times(self, times):
        for i, t in enumerate(times):
            self.content.entries[i].time = int(1000 * t)

    def build(self):
        data = structs.AnlzTag.build(self.struct)
        if self.struct.content.entry_count == 0:
            data = data[: self.struct.len_tag]

        len_data = len(data)
        if len_data != self.struct.len_tag:
            raise BuildTagLengthError(self.struct, len_data)
        return data


class PCOBAnlzTag(AbstractAnlzTag):
    """Cue list struct handler."""

    type = "PCOB"
    name = "cue_list"
    LEN_HEADER = 24


class PCO2AnlzTag(AbstractAnlzTag):
    """Extended (nxs2) cue list struct handler."""

    type = "PCO2"
    name = "cue_list2"
    LEN_HEADER = 20


class PPTHAnlzTag(AbstractAnlzTag):
    """Path struct handler."""

    type = "PPTH"
    name = "path"
    LEN_HEADER = 16

    @property
    def path(self):
        return self.content.path

    def get(self):
        return self.content.path

    def set(self, path):
        path = path.replace("\\", "/")
        len_path = len(path.encode("utf-16-be")) + 2
        self.content.path = path
        self.content.len_path = len_path

    def update_len(self):
        self.struct.len_tag = self.struct.len_header + self.content.len_path


class PVBRAnlzTag(AbstractAnlzTag):
    """VBR struct handler."""

    type = "PVBR"
    name = "vbr"
    LEN_HEADER = 16
    LEN_TAG = 1620

    def get(self):
        return np.array(self.content.idx)


class PSSIAnlzTag(AbstractAnlzTag):
    """Song structure struct handler."""

    type = "PSSI"
    name = "structure"
    LEN_HEADER = 32


class PWAVAnlzTag(AbstractAnlzTag):
    """Waveform preview struct handler."""

    type = "PWAV"
    name = "wf_preview"
    LEN_HEADER = 20

    def get(self):
        return _parse_wf_preview(self.content)


class PWV2AnlzTag(AbstractAnlzTag):
    """Tiny waveform preview struct handler."""

    type = "PWV2"
    name = "wf_tiny_preview"
    LEN_HEADER = 20

    def get(self):
        return _parse_wf_preview(self.content)


class PWV3AnlzTag(AbstractAnlzTag):
    """Waveform detail struct handler."""

    type = "PWV3"
    name = "wf_detail"
    LEN_HEADER = 24

    def get(self):
        n = len(self.content.entries)
        wf = np.zeros(n)
        col = np.zeros(n)
        for i in range(n):
            data = self.content.entries[i]
            wf[i] = data & 0x1F
            col[i] = data >> 5
        return wf, col


class PWV4AnlzTag(AbstractAnlzTag):
    """Waveform color preview struct handler."""

    type = "PWV4"
    name = "wf_color"
    LEN_HEADER = 24

    def get(self):
        num_entries = self.content.len_entries
        data = self.content.entries
        ws, hs = 1, 1
        w = int(num_entries / ws)
        # parse color and blue waveforms
        col_color = np.zeros((num_entries, 2, 3), dtype=np.int64)
        col_blues = np.zeros((num_entries, 2, 3), dtype=np.int64)
        heights = np.zeros((num_entries, 2), dtype=np.int64)
        for x in range(w):
            # d0 = data[x * ws * 6 + 0]  # unknown?
            d1 = data[x * ws * 6 + 1]  # some kind of luminance boost?
            d2 = data[x * ws * 6 + 2] & 0x7F  # inverse intensity for blue waveform
            d3 = data[x * ws * 6 + 3] & 0x7F  # red
            d4 = data[x * ws * 6 + 4] & 0x7F  # green
            d5 = data[x * ws * 6 + 5] & 0x7F  # blue and height of front waveform
            bh = int(max(d2, d3, d4) / hs)  # back height is max of d3, d4 probably d2?
            fh = int(d5 / hs)  # front height is d5
            fl = 32  # front luminosity increase (arbitrary)
            # waveform heights
            heights[x] = fh, bh
            # color waveform
            col = np.array([d3, d4, d5]) * (d1 / 127)
            col_color[x] = col, col + fl
            # blue waveform
            col = 95 - d2 * np.array([1.0, 0.5, 0.25])
            col_blues[x] = col, col + fl
        return heights, col_color, col_blues


class PWV5AnlzTag(AbstractAnlzTag):
    """Waveform color detail struct handler."""

    type = "PWV5"
    name = "wf_color_detail"
    LEN_HEADER = 24

    def get(self):
        """Parse the Waveform Color Detail Tag (PWV5)

        The format of the entries is:

          f  e  d  c  b  a  9  8  7  6  5  4  3  2  1  0
        │   red  │  green │  blue  │    height    │00 00│
        """
        rmask = 0xE000  # 111 000 000 00000 00
        gmask = 0x1C00  # 000 111 000 00000 00
        bmask = 0x380  # 000 000 111 00000 00
        hmask = 0x7C  # 000 000 000 11111 00

        n = self.content.len_entries
        heights = np.zeros(n, dtype=np.int64)
        colors = np.zeros((n, 3), dtype=np.int64)
        for i, x in enumerate(self.content.entries):
            red = (x & rmask) >> 12
            green = (x & gmask) >> 10
            blue = (x & bmask) >> 7
            heights[i] = (x & hmask) >> 2
            colors[i] = red, green, blue
        # Normalize heights to 1:
        heights = heights / 31
        return heights, colors


class PWV6AnlzTag(AbstractAnlzTag):
    """PWV6 struct handler."""

    type = "PWV6"
    name = type
    LEN_HEADER = 20


class PWV7AnlzTag(AbstractAnlzTag):
    """PWV7 struct handler."""

    type = "PWV7"
    name = type
    LEN_HEADER = 24


class PWVCAnlzTag(AbstractAnlzTag):
    """PWVC struct handler."""

    type = "PWVC"
    name = type
    LEN_HEADER = 14


TAGS = {
    "PQTZ": PQTZAnlzTag,
    "PQT2": PQT2AnlzTag,
    "PCOB": PCOBAnlzTag,  # seen in both DAT and EXT files
    "PCO2": PCO2AnlzTag,  # seen in EXT files
    "PPTH": PPTHAnlzTag,
    "PVBR": PVBRAnlzTag,
    "PSSI": PSSIAnlzTag,  # seen in EXT files
    "PWAV": PWAVAnlzTag,
    "PWV2": PWV2AnlzTag,
    "PWV3": PWV3AnlzTag,  # seen in EXT files
    "PWV4": PWV4AnlzTag,  # seen in EXT files
    "PWV5": PWV5AnlzTag,  # seen in EXT files
    "PWV6": PWV6AnlzTag,  # seen in 2EX files
    "PWV7": PWV7AnlzTag,  # seen in 2EX files
    "PWVC": PWVCAnlzTag,  # seen in 2EX files
}



# coding: utf-8
#
# This code is part of pyrekordbox.
#
# Copyright (c) 2022, Dylan Jones

import os
from pyrekordbox.mysettings import read_mysetting_file, FILES
from pyrekordbox.mysettings.file import compute_checksum

TEST_ROOT = os.path.join(os.path.dirname(os.path.dirname(__file__)), ".testdata")


def _read_setting(type_, key, value):
    file = type_.upper() + ".DAT"
    root = os.path.join(TEST_ROOT, "mysettings", type_)
    path = os.path.join(root, key, value, file)
    if not os.path.exists(path):
        raise FileNotFoundError(path)
    file = read_mysetting_file(path)
    return file[key]


def _read_mysetting(key, value):
    return _read_setting("mysetting", key, value)


def _read_mysetting2(key, value):
    return _read_setting("mysetting2", key, value)


def _read_djmysetting(key, value):
    return _read_setting("djmmysetting", key, value)


def _get_values(type_, key):
    root = os.path.join(TEST_ROOT, "mysettings", type_, key)
    return os.listdir(root)


def get_mysetting_values(key):
    return _get_values("mysetting", key)


def get_mysetting2_values(key):
    return _get_values("mysetting2", key)


def get_djmmysetting_values(key):
    return _get_values("djmmysetting", key)


# -- MySetting -------------------------------------------------------------------------


def test_mysetting_on_air_display():
    key = "on_air_display"
    for expected in get_mysetting_values(key):
        assert _read_mysetting(key, expected) == expected


def test_mysetting_lcd_brightness():
    key = "lcd_brightness"
    for expected in get_mysetting_values(key):
        assert _read_mysetting(key, expected) == expected


def test_mysetting_quantize():
    key = "quantize"
    for expected in get_mysetting_values(key):
        assert _read_mysetting(key, expected) == expected


def test_mysetting_auto_cue_level():
    key = "auto_cue_level"
    for expected in get_mysetting_values(key):
        assert _read_mysetting(key, expected) == expected


def test_mysetting_language():
    key = "language"
    for expected in get_mysetting_values(key):
        assert _read_mysetting(key, expected) == expected


def test_mysetting_jog_ring_brightness():
    key = "jog_ring_brightness"
    for expected in get_mysetting_values(key):
        assert _read_mysetting(key, expected) == expected


def test_mysetting_jog_ring_indicator():
    key = "jog_ring_indicator"
    for expected in get_mysetting_values(key):
        assert _read_mysetting(key, expected) == expected


def test_mysetting_slip_flashing():
    key = "slip_flashing"
    for expected in get_mysetting_values(key):
        assert _read_mysetting(key, expected) == expected


def test_mysetting_eject_lock():
    key = "eject_lock"
    for expected in get_mysetting_values(key):
        assert _read_mysetting(key, expected) == expected


def test_mysetting_disc_slot_illumination():
    key = "disc_slot_illumination"
    for expected in get_mysetting_values(key):
        assert _read_mysetting(key, expected) == expected


def test_mysetting_sync():
    key = "sync"
    for expected in get_mysetting_values(key):
        assert _read_mysetting(key, expected) == expected


def test_mysetting_play_mode():
    key = "play_mode"
    for expected in get_mysetting_values(key):
        assert _read_mysetting(key, expected) == expected


def test_mysetting_quantize_beat_value():
    key = "quantize_beat_value"
    for expected in get_mysetting_values(key):
        assert _read_mysetting(key, expected) == expected


def test_mysetting_hotcue_autoload():
    key = "hotcue_autoload"
    for expected in get_mysetting_values(key):
        assert _read_mysetting(key, expected) == expected


def test_mysetting_hotcue_color():
    key = "hotcue_color"
    for expected in get_mysetting_values(key):
        assert _read_mysetting(key, expected) == expected


def test_mysetting_needle_lock():
    key = "needle_lock"
    for expected in get_mysetting_values(key):
        assert _read_mysetting(key, expected) == expected


def test_mysetting_time_mode():
    key = "time_mode"
    for expected in get_mysetting_values(key):
        assert _read_mysetting(key, expected) == expected


def test_mysetting_jog_mode():
    key = "jog_mode"
    for expected in get_mysetting_values(key):
        assert _read_mysetting(key, expected) == expected


def test_mysetting_auto_cue():
    key = "auto_cue"
    for expected in get_mysetting_values(key):
        assert _read_mysetting(key, expected) == expected


def test_mysetting_master_tempo():
    key = "master_tempo"
    for expected in get_mysetting_values(key):
        assert _read_mysetting(key, expected) == expected


def test_mysetting_tempo_range():
    key = "tempo_range"
    for expected in get_mysetting_values(key):
        assert _read_mysetting(key, expected) == expected


def test_mysetting_phase_meter():
    key = "phase_meter"
    for expected in get_mysetting_values(key):
        assert _read_mysetting(key, expected) == expected


# -- MySetting2 ------------------------------------------------------------------------


def test_mysetting2_vinyl_speed_adjust():
    key = "vinyl_speed_adjust"
    for expected in get_mysetting2_values(key):
        assert _read_mysetting2(key, expected) == expected


def test_mysetting2_jog_display_mode():
    key = "jog_display_mode"
    for expected in get_mysetting2_values(key):
        assert _read_mysetting2(key, expected) == expected


def test_mysetting2_pad_button_brightness():
    key = "pad_button_brightness"
    for expected in get_mysetting2_values(key):
        assert _read_mysetting2(key, expected) == expected


def test_mysetting2_jog_lcd_brightness():
    key = "jog_lcd_brightness"
    for expected in get_mysetting2_values(key):
        assert _read_mysetting2(key, expected) == expected


def test_mysetting2_waveform_divisions():
    key = "waveform_divisions"
    for expected in get_mysetting2_values(key):
        assert _read_mysetting2(key, expected) == expected


def test_mysetting2_waveform():
    key = "waveform"
    for expected in get_mysetting2_values(key):
        assert _read_mysetting2(key, expected) == expected


def test_mysetting2_beat_jump_beat_value():
    key = "beat_jump_beat_value"
    for expected in get_mysetting2_values(key):
        assert _read_mysetting2(key, expected) == expected


# -- DjmMySetting ----------------------------------------------------------------------


def test_mysetting2_channel_fader_curve():
    key = "channel_fader_curve"
    for expected in get_djmmysetting_values(key):
        assert _read_djmysetting(key, expected) == expected


def test_mysetting2_crossfader_curve():
    key = "cross_fader_curve"
    for expected in get_djmmysetting_values(key):
        assert _read_djmysetting(key, expected) == expected


def test_mysetting2_headphones_pre_eq():
    key = "headphones_pre_eq"
    for expected in get_djmmysetting_values(key):
        assert _read_djmysetting(key, expected) == expected


def test_mysetting2_headphones_mono_split():
    key = "headphones_mono_split"
    for expected in get_djmmysetting_values(key):
        assert _read_djmysetting(key, expected) == expected


def test_mysetting2_beat_fx_quantize():
    key = "beat_fx_quantize"
    for expected in get_djmmysetting_values(key):
        assert _read_djmysetting(key, expected) == expected


def test_mysetting2_mic_low_cut():
    key = "mic_low_cut"
    for expected in get_djmmysetting_values(key):
        assert _read_djmysetting(key, expected) == expected


def test_mysetting2_talk_over_mode():
    key = "talk_over_mode"
    for expected in get_djmmysetting_values(key):
        assert _read_djmysetting(key, expected) == expected


def test_mysetting2_talk_over_level():
    key = "talk_over_level"
    for expected in get_djmmysetting_values(key):
        assert _read_djmysetting(key, expected) == expected


def test_mysetting2_midi_channel():
    key = "midi_channel"
    for expected in get_djmmysetting_values(key):
        assert _read_djmysetting(key, expected) == expected


def test_mysetting2_midi_button_type():
    key = "midi_button_type"
    for expected in get_djmmysetting_values(key):
        assert _read_djmysetting(key, expected) == expected


def test_mysetting2_display_brightness():
    key = "display_brightness"
    for expected in get_djmmysetting_values(key):
        assert _read_djmysetting(key, expected) == expected


def test_mysetting2_indicator_brightness():
    key = "indicator_brightness"
    for expected in get_djmmysetting_values(key):
        assert _read_djmysetting(key, expected) == expected


def test_mysetting2_channel_fader_curve_long():
    key = "channel_fader_curve_long"
    for expected in get_djmmysetting_values(key):
        assert _read_djmysetting(key, expected) == expected


# ======================================================================================


def test_mysetting_checksum():
    file_types = "MYSETTING.DAT", "MYSETTING2.DAT", "DJMMYSETTING.DAT"

    root = os.path.join(TEST_ROOT, "mysettings")
    for root, _, files in os.walk(root):
        for name in files:
            if name not in file_types:
                continue
            path = os.path.join(root, name)
            sett = read_mysetting_file(path)
            data = sett.build()
            checksum = compute_checksum(data, sett.struct)
            assert checksum == sett.parsed.checksum


def test_default_build():
    file_types = "MYSETTING.DAT", "MYSETTING2.DAT", "DJMMYSETTING.DAT"
    for key in file_types:
        obj = FILES[key]()
        data = obj.build()
        # See if data is parsable
        obj.parse(data)



# coding: utf-8
#
# This code is part of pyrekordbox.
#
# Copyright (c) 2022, Dylan Jones

import os
from pyrekordbox import RekordboxXml

TEST_ROOT = os.path.join(os.path.dirname(os.path.dirname(__file__)), ".testdata")


def test_parse_xml_tracks_v5():
    path = os.path.join(TEST_ROOT, "rekordbox 5", "database.xml")
    xml = RekordboxXml(path)
    assert os.path.exists(path)

    assert xml.num_tracks == 6

    track = xml.get_track(1)
    assert track.name == "NOISE"
    assert len(track.tempos) == 0
    assert len(track.cue_points()) == 0

    track = xml.get_track(5)

    tempo = track.tempos[0]
    assert tempo.bpm == 128.0
    assert tempo.inizio == 0.025
    assert tempo.metro == "4/4"
    assert tempo.battito == 1

    positions = track.cue_points()
    assert len(positions) == 4
    assert positions[0].name == ""
    assert positions[0].type == 0
    assert positions[0].start == 0.025
    assert positions[0].num == -1



# coding: utf-8
#
# This code is part of pyrekordbox.
#
# Copyright (c) 2022, Dylan Jones

import os
import re
from . import structs
from .file import AnlzFile

RE_ANLZ = re.compile("ANLZ[0-9]{4}.(DAT|EXT|2EX)")


def get_anlz_paths(root):
    paths = {"DAT": None, "EXT": None, "2EX": None}
    for fname in os.listdir(root):
        if RE_ANLZ.match(fname):
            ext = os.path.splitext(fname)[1][1:]
            paths[ext] = os.path.join(root, fname)
    return paths


def walk_anlz_dirs(path):
    path = os.path.abspath(path)
    for root, _, names in os.walk(path):
        for fname in names:
            if RE_ANLZ.match(fname):
                yield root


def walk_anlz_paths(path):
    path = os.path.abspath(path)
    for root, _, names in os.walk(path):
        files = dict()
        for fname in names:
            if RE_ANLZ.match(fname):
                ext = os.path.splitext(fname)[1]
                files[ext[1:].upper()] = os.path.join(root, fname)
        if files:
            yield root, files


def read_anlz_files(root=""):
    paths = get_anlz_paths(root)
    return {key: AnlzFile.parse_file(path) for key, path in paths.items()}
